#include "display.h"
#include <stdarg.h>
#include <string.h>

rtems_unsigned8 _scr_buf[68]={
	0,0,0,0, 0,0,0,0,	0,0,0,0, 0,0,0,0,
	0,0,0,0, 0,0,0,0,	0,0,0,0, 0,0,0,0,
	0,0,0,0, 0,0,0,0,	0,0,0,0, 0,0,0,0,
	0,0,0,0, 0,0,0,0,	0,0,0,0, 0,0,0,0,
	0,0,0,0
};

rtems_unsigned8 *screen=_scr_buf;
rtems_unsigned8 *actualscreen=_scr_buf+34;
rtems_unsigned8 cursor_chr=0;
int volatile cursor_ofs=0;
int volatile cursor_enabled=0;
int volatile cursor_displayed=0;
int volatile panel_input_disabled=0;

rtems_unsigned8* chr_set[9]={NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};

int _screen_is_valid=1;

int cursor_start_line = 0;
int cursor_end_line   = 7;

rtems_id disp_mutex;

rtems_interval displaying_message = 0;

extern int volatile panel;

rtems_unsigned8 msgbuf[33];

int volatile blink=0;
//char old_cc=0;


__inline void _disp_cursor() {
	char cstr[2]={CHR_CURSOR,0};
	if(cursor_enabled&&!cursor_displayed&&blink&&!displaying_message) {
		if(cursor_chr!=actualscreen[cursor_ofs]) {
			int i;
			cursor_chr=actualscreen[cursor_ofs];
			for(i=0;i<cursor_start_line;i++) {
				*(rtems_unsigned8*)(cursor_bitmap+i) = *(ASCII_BITMAP(cursor_chr)+i);
			}
			for(;i<=cursor_end_line;i++) {
				*(rtems_unsigned8*)(cursor_bitmap+i) = 0x1F^*(ASCII_BITMAP(cursor_chr)+i);
			}
			for(;i<8;i++) {
				*(rtems_unsigned8*)(cursor_bitmap+i) = *(ASCII_BITMAP(cursor_chr)+i);
			}
			if (!panel_out_lcd_redefine(panel, CHR_CURSOR, cursor_bitmap))
				TRACE("ERROR: cannot redefine cursor symbol.\n");
		}
		panel_out_lcd_print(panel,cursor_ofs>>4,cursor_ofs&0xF,cstr);
		cursor_displayed=1;
	}
}


__inline void _hide_cursor() {
	char cstr[2]={0,0};
	if(cursor_enabled&&cursor_displayed&&!blink&&!displaying_message) {
		cstr[0]=actualscreen[cursor_ofs];
		panel_out_lcd_print(panel,cursor_ofs>>4,cursor_ofs&0xF,cstr);
		cursor_displayed=0;
	}
}


__inline void _refresh() {
	int begin=0,end=0;
	rtems_unsigned8*source;
	rtems_interval date;
	rtems_unsigned8 backup;
	rtems_clock_get(RTEMS_CLOCK_GET_TICKS_SINCE_BOOT,&date);
	if(date>=displaying_message+MESSAGE_HOLD) {
		source=screen;
		displaying_message=0;
//		TRACE("display : screen %u:%u\n",date,displaying_message);
	} else {
		source=msgbuf;
//		TRACE("display : message %u:%u\n",date,displaying_message);
	}
	/*TRACE("@@@###+++            REFRESH             +++###@@@\n",begin,end);*/
	_hide_cursor();
	while(end<16) {
		/*TRACE("@@@###+++ do screen refresh loop (%i/%i) +++###@@@\n",begin,end);*/
		while(begin<16&&source[begin]==actualscreen[begin]) ++begin;
		end=begin;
		if(begin<16) {
			/*TRACE("screen diff %2.2i : %X / %X\n",begin,screen[begin],actualscreen[begin]);*/
			actualscreen[begin]=source[begin];
			++end;
			while(end<16&&source[end]!=actualscreen[end]) {
				/*TRACE("screen diff %2.2i : %X / %X\n",end,screen[end],actualscreen[end]);*/
				actualscreen[end]=source[end];
				++end;
			}
			backup=actualscreen[end];
			actualscreen[end]=0;
			/*TRACE("@@@###+++ update screen (%i/%i) +++###@@@\n",begin,end);*/
			panel_out_lcd_print(panel,0,begin,actualscreen+begin);
			actualscreen[end]=backup;
		}
	};
	begin=16;
	while(end<32) {
		/*TRACE("@@@###+++ do screen refresh loop (%i/%i) +++###@@@\n",begin,end);*/
		while(begin<32&&source[begin]==actualscreen[begin]) ++begin;
		end=begin;
		if(begin<32) {
			/*TRACE("screen diff %2.2i : %X / %X\n",begin,screen[begin],actualscreen[begin]);*/
			actualscreen[begin]=source[begin];
			++end;
			while(end<32&&source[end]!=actualscreen[end]) {
				/*TRACE("screen diff %2.2i : %X / %X\n",end,screen[end],actualscreen[end]);*/
				actualscreen[end]=source[end];
				++end;
			}
			backup=actualscreen[end];
			actualscreen[end]=0;
			/*TRACE("@@@###+++ update screen (%i/%i) +++###@@@\n",begin,end);*/
			panel_out_lcd_print(panel,1,begin&0xF,actualscreen+begin);
			actualscreen[end]=backup;
		}
	};
	_disp_cursor();
	_screen_is_valid=1;
}


__inline void _invalidate_screen() {
	int i;
	for(i=0;i<32;i++) {
		actualscreen[i]=0;
	}
	_screen_is_valid=0;
}

__inline void _redraw() {
	_invalidate_screen();
	_refresh();
}


__inline void _display(int x,int y,char*text) {
	int ofs=x+(y<<4);
	int len=16-x;
	if((y&~1)||(x&~15))
		return;
	while(*text&&len--) screen[ofs++]=*text++;
}


__inline void _disable_cursor() {
	_hide_cursor();
	cursor_enabled=0;
}


__inline void _enable_cursor() {
	cursor_enabled=1;
}


__inline void _set_cursor(int x,int y) {
	static int volatile oldco=32; /* out of screen */
	int b;
	if((y&~1)||(x&~15))
		return;
	_enable_cursor();
	b=blink;
	blink=0;
	_hide_cursor();
	cursor_ofs=x+(y<<4);
	if(oldco!=cursor_ofs&&cursor_displayed) {
		screen[oldco]=cursor_chr;
		actualscreen[oldco]=0;
		oldco=cursor_ofs;
	}
	//cursor_chr=actualscreen[cursor_ofs];
	blink=b;
	_disp_cursor();
	/*screen[cursor_ofs]=CHR_CURSOR;*/
}


__inline void _clear() {
	_hide_cursor();
	_disable_cursor();
	memset(screen,' ',32);
}


__inline void _set_chr(int c,rtems_unsigned8 const* bm) {
	int i;
	/*TRACE("SET CHR :::::::: %u => %p / %p\n",c,chr_set[c],bm);*/
	if(/*_screen_is_valid&&*/chr_set[c]!=bm) {
		/*TRACE("                 CHANGE CHR !\n");*/
		for(i=0;i<8;i++) {
			*(rtems_unsigned8*)(ASCII_BITMAP(c)+i)=*(bm+i);
		}
		if (!panel_out_lcd_redefine(panel, c, bm))
			TRACE("ERROR: cannot redefine cursor symbol.\n");
		/* invalidate characters on screen */
		for(i=0;i<32;i++) if(actualscreen[i]==c) actualscreen[i]=0;
		chr_set[c]=(rtems_unsigned8*)bm;
	}
}


__inline void _raw_set_chr(int c,rtems_unsigned8 const* bm) {
	if (panel_out_lcd_redefine(panel, c, bm))
		chr_set[c]=(rtems_unsigned8*)bm;
	else
		TRACE("ERROR: cannot redefine cursor symbol.\n");
}

























void redraw() {
	disp_lock();
	_redraw();
	disp_unlock();
}


void clear() {
	disp_lock();
	_clear();
	disp_unlock();
}


void display(int x,int y,char*text,...) {
	va_list args;
	int ofs=(y<<4)+x;
	va_start(args,text);
	vsnprintf(msgbuf,33-ofs,text,args);
	/*TRACE("msgbuf :::::::::::: '%s'\n",msgbuf);*/
	va_end(args);
	disp_lock();
	_display(x,y,msgbuf);
	disp_unlock();
}


void set_cursor(int x,int y) {
	disp_lock();
	_set_cursor(x,y);
	disp_unlock();
}


void disable_cursor() {
	disp_lock();
	_disable_cursor();
	disp_unlock();
}


void enable_cursor() {
	disp_lock();
	if(!cursor_enabled) {
		screen[cursor_ofs]=CHR_CURSOR;
		cursor_enabled=1;
	}
	disp_unlock();
}


void set_chr(int c,rtems_unsigned8 const* bm) {
	disp_lock();
	_set_chr(c,bm);
	disp_unlock();
}


void raw_set_chr(int c,rtems_unsigned8 const* bm) {
	disp_lock();
	_raw_set_chr(c,bm);
	disp_unlock();
}


void refresh() {
	disp_lock();
	_refresh();
	disp_unlock();
}


void invalidate_screen() {
	disp_lock();
	_invalidate_screen();
	disp_unlock();
}



void do_leds();

#define BLINK_LED_PERIOD 100
#define BLINK_CURSOR_PERIOD 500
#define BLINK_RATIO (BLINK_CURSOR_PERIOD/BLINK_LED_PERIOD)

static rtems_task panel_blink_task(rtems_task_argument argument) {
	rtems_id period;
	int i=10;

	rtems_rate_monotonic_create(rtems_build_name('B','L','K','P'),&period);

	while(TRUE) {
		rtems_rate_monotonic_period(period,BLINK_LED_PERIOD);
		/*TRACE("panel_blink_task one period\n");*/
		++i;
		i%=BLINK_RATIO;
		do_leds();
		//TRACE("!i %i cursor_enabled %i !panel_input_disabled %i\n",!i,cursor_enabled,!panel_input_disabled);
		if((!i)&&cursor_enabled&&!displaying_message) {
			disp_lock();
			if(cursor_displayed) {
				blink=0;
				_hide_cursor();
			} else {
				blink=1;
				_disp_cursor();
			}
			disp_unlock();
		}
		/*TRACE("panel_blink_task period done\n");*/
	}
	rtems_task_delete(RTEMS_SELF);
}


void do_cursor() {
	static int i;
	++i;
	i%=12;
	//TRACE("!i %i cursor_enabled %i !panel_input_disabled %i\n",!i,cursor_enabled,!panel_input_disabled);
	if((!i)&&cursor_enabled&&!displaying_message) {
		disp_lock();
		if(cursor_displayed) {
			blink=0;
			_hide_cursor();
		} else {
			blink=1;
			_disp_cursor();
		}
		disp_unlock();
	}
}



int disp_init() {
	rtems_id task_id;
	int i,ret;
	ret=rtems_semaphore_create(rtems_build_name('D','M','T','X'),1,
					0,0,&disp_mutex);
	/*TRACE("\n@ # + # @      Mutex %X\n\n",task_id);*/
	/*rtems_semaphore_flush(disp_mutex);*/
/*	display(0,0,"Test :D \\o/  \001\002\003");
	set_chr(4,BOLD_BITMAP('b'));
	set_chr(5,BOLD_BITMAP('l'));
	set_chr(6,BOLD_BITMAP('0'));
	set_chr(7,BOLD_BITMAP('b'));
	display(0,1,"\004\005\006\007lablablabla!");*/
	for(i=0;i<8;i++) raw_set_chr(i,ASCII_BITMAP(0));
/*	return	RTEMS_SUCCESSFUL==rtems_task_create(rtems_build_name('B','L','N','K'),
						PANEL_PRIO-1,RTEMS_MINIMUM_STACK_SIZE,
						RTEMS_DEFAULT_ATTRIBUTES,
						RTEMS_DEFAULT_MODES,&task_id)
		&&
		RTEMS_SUCCESSFUL==rtems_task_start(task_id, panel_blink_task,
						(rtems_task_argument)0);
*/
	return 1;
}



//static rtems_task display_message_task(rtems_task_argument argument) {
//	static int volatile reentrency_count=0;
//	char*text=(char*)argument;
//	int len=strlen(text);
//	int ce_backup=cursor_enabled;
//	++reentrency_count;
/*//	TRACE("\n\nSTART DISPLAY MESSAGE %s\n",text);*/
//	disp_lock();
//	panel_input_disabled=1;
//	if(ce_backup) _disable_cursor();
//	panel_out_lcd_print(panel,0,0,text);
//	if(len<16) {
//		panel_out_lcd_print(panel,0,len,"                ");
//		panel_out_lcd_print(panel,1,0,"                ");
//	} else {
//		panel_out_lcd_print(panel,1,0,text+16);
//		if(len<32)
//			panel_out_lcd_print(panel,1,len-16,"                ");
//	}
//	rtems_task_wake_after(MESSAGE_HOLD);
//	_invalidate_screen();
//	disp_unlock();
//	invalidate_screen();
//	panel_input_disabled=0;
//	if(reentrency_count==1)
//		rtems_task_wake_after(MESSAGE_DURATION-MESSAGE_HOLD);
//	--reentrency_count;
//	if(!reentrency_count) {
//		if(ce_backup) _enable_cursor();
//		refresh();
//	}
//	/*rtems_task_wake_after(MESSAGE_HALFPERIOD);*/
/*//	TRACE("END DISPLAY MESSAGE\n\n\n");*/
//	rtems_task_delete(RTEMS_SELF);
//}


extern int volatile mode;

void display_message(char*text,...) {
	va_list args;
//	static char k='9';
//	rtems_id task_id;
//	int ret;
	if(mode) return;
	disp_lock();
	va_start(args,text);
	//memset(msgbuf,32,32);
	vsnprintf(msgbuf,33,text,args);
	va_end(args);
//	TRACE("msgbuf :::::::::::: '%s'\n",msgbuf);
	_disable_cursor();
	rtems_clock_get(RTEMS_CLOCK_GET_TICKS_SINCE_BOOT,&displaying_message);
	_redraw();
	disp_unlock();
	//if(k=='9') k=0; else ++k;
//	ret=rtems_task_create(rtems_build_name('M','S','G',k),
//				PANEL_PRIO-2,RTEMS_MINIMUM_STACK_SIZE,
//				RTEMS_DEFAULT_ATTRIBUTES,
//				RTEMS_DEFAULT_MODES,&task_id);
/*	TRACE("MSG TASK CREATE RETURNED %i\n",ret);*/
//	rtems_task_start(task_id, display_message_task,(rtems_task_argument)msgbuf);
/*	TRACE("MSG TASK START RETURNED %i\n",ret);*/
}

