/*******************************************************************************
 * C H A M E L E O N   DSP C file                                              *
 *******************************************************************************
 * Copyright (C) 2001-2002 Soundart                                            *
 * www.soundart-hot.com                                                        *
 * support@soundart-hot.com                                                    *
 ******************************************************************************/

#include <stdlib.h>
#include <dsp56303.h>

static int *m_pActiveBufferDAC;
static int *m_pActiveBufferADC;

void ConfigAudioDMA(int nBufferSize, int pBufferDAC[2][], int pBufferADC[2][])
{
	/**********************************************************************
	 * Disable ESSI0 transmitter and receiver 
	 *********************************************************************/
	CRB0.B.TE0 = 0;
	CRB0.B.SRE = 0;

	/**********************************************************************
	 * Disable DMA Channels 0 and 1
	 *********************************************************************/
	DCR0.B.DE = 0;
	DCR1.B.DE = 0;

	/**********************************************************************
	 * DOR0: DMA OFFSET REGISTER 0 (DSP56300FM page 8-3)
	 * Offset: -(2*BUFFER_SIZE-1)
	 *********************************************************************/
	DOR0 = -((2*2)*nBufferSize-1);

	/**********************************************************************
	 * DCO0: DMA COUNTER CHANNEL 0 (DSP56300FM page 8-5)
	 * DCOL: 0
	 * DCOH: 2*BUFFER_SIZE-1
	 *********************************************************************/
	DCO0 = (2*2)*nBufferSize-1;

	/**********************************************************************
	 * DSR0: DMA SOURCE ADDRESS REGISTER CHANNEL 0 (DSP56300FM page 8-2)
	 *********************************************************************/
	DSR0 = (int *) pBufferDAC;

	/**********************************************************************
	 * DDR0: DMA DESTINATION ADDRESS REGISTER CHANNEL 0 (DSP56300FM page 8-2)
	 *********************************************************************/
	DDR0 = (int *) &TX00;

	/**********************************************************************
	 * DCR0: DMA CONTROL REGISTER CHANNEL 0 (DSP56300FM page 8-8)
	 * DMA Source Space: X mem				(00)
	 * DMA Destination Space: X mem				(00)
	 * DMA Address Mode: Mode B, DOR0, src 2D, dst no update(100000)
	 * DMA Three Dimensional Mode: Off			(0)
	 * DMA Request Source: ESSI0Tx				(01011)
	 * DMA Continuous Mode					(0)
	 * DMA Channel Priority: 3				(11)
	 * DMA Transfer Mode: word transfer, no clear DE	(101)
	 * DMA Interrupt Enable: Off				(0)
	 * DMA Channel Enable: Off				(0)
	 **********************************************************************
	 * 0 0 101 11 0 01011 0 100000 00 00 = 0x2E5A00
	 *********************************************************************/
	DCR0.I = 0x2E5A00;

	/**********************************************************************
	 * DOR1: DMA OFFSET REGISTER 1 (DSP56300FM page 8-3)
	 * Offset: -(2*BUFFER_SIZE-1)
	 *********************************************************************/
	DOR1 = -((2*2)*nBufferSize-1);

	/**********************************************************************
	 * DCO1: DMA COUNTER CHANNEL 1 (DSP56300FM page 8-5)
	 * DCOL: 0
	 * DCOH: 2*BUFFER_SIZE-1
	 *********************************************************************/
	DCO1 = (2*2)*nBufferSize-1;

	/**********************************************************************
	 * DDR1: DMA DESTINATION ADDRESS REGISTER CHANNEL 1 (DSP56300FM page 8-2)
	 *********************************************************************/
	DDR1 = (int *) pBufferADC;

	/**********************************************************************
	 * DSR1: DMA SOURCE ADDRESS REGISTER CHANNEL 1 (DSP56300FM page 8-2)
	 *********************************************************************/
	DSR1 = (int *) &RX0;

	/**********************************************************************
	 * DCR1: DMA CONTROL REGISTER CHANNEL 1 (DSP56300FM page 8-8)
	 * DMA Source Space: X mem				(00)
	 * DMA Destination Space:X mem				(00)
	 * DMA Address Mode: Mode B, DOR1, dst 2D, src no update(001100)
	 * DMA Three Dimensional Mode: Off			(0)
	 * DMA Request Source: ESSI0Rx				(01010)
	 * DMA Continuous Mode					(0)
	 * DMA Channel Priority: 3				(11)
	 * DMA Transfer Mode: word transfer, no clear DE	(101)
	 * DMA Interrupt Enable: Off				(0)
	 * DMA Channel Enable: Off				(0)
	 **********************************************************************
	 * 0 0 101 11 0 01010 0 001100 00 00 = 0x2E50C0
	 *********************************************************************/
	DCR1.I = 0x2E50C0;

	/**********************************************************************
	 * Initialize private data
	 *********************************************************************/
	m_pActiveBufferDAC = ((int *) pBufferDAC) + 2*nBufferSize;
	m_pActiveBufferADC = ((int *) pBufferADC) + 2*nBufferSize;

	{
		/* Empty DAC buffers */
		int *pDAC = (int *) pBufferDAC;
		int i;
		
		for (i=0; i<(2*2)*nBufferSize ;i++)
			*pDAC++ = 0;
	}
	
	/**********************************************************************
	 * Enable ESSI0 transmitter and receiver 
	 *********************************************************************/
	TX00 = 0;
	CRB0.B.TE0 = 1;
	CRB0.B.SRE = 1;
	
	while (!SSISR0.B.TDE)
		continue;
	
	if (!SSISR0.B.TFS)	/* left channel */
	{
		TX00 = 0;
		while (!SSISR0.B.TDE)
			continue;
	}
	TX00 = 0;
		
	/**********************************************************************
	 * Enable DMA Channel 0 (ESSI0 TX)
	 *********************************************************************/
	DCR0.B.DE = 1;

	RX0;
	while (!SSISR0.B.RDF)
		continue;
		
	if (!SSISR0.B.RFS)
	{
		RX0;
		while (!SSISR0.B.RDF)
			continue;
	}

	/**********************************************************************
	 * Enable DMA Channel 1 (ESSI0 RX)
	 *********************************************************************/
	DCR1.B.DE = 1;
}

void ChangeActiveBufferDMA(int nBufferSize, int pBufferDAC[2][], int pBufferADC[2][])
{
	if (m_pActiveBufferDAC == (((int *) pBufferDAC) + 2*nBufferSize))
	{
		while (DSR0 < (int *) (((int *) pBufferDAC) + 2*nBufferSize))
			continue;
		
		m_pActiveBufferDAC = (int *) pBufferDAC;
		m_pActiveBufferADC = (int *) pBufferADC;
	}
	else
	{
		while (DSR0 >= (int *) (((int *) pBufferDAC) + 2*nBufferSize))
			continue;
		
		m_pActiveBufferDAC = ((int *) pBufferDAC) + 2*nBufferSize;
		m_pActiveBufferADC = ((int *) pBufferADC) + 2*nBufferSize;
	}
}

int *GetActiveBufferDAC(void)
{
	return m_pActiveBufferDAC;
}

int *GetActiveBufferADC(void)
{
	return m_pActiveBufferADC;
}

int GetUsageBufferDMA(int nBufferSize, int pBufferDAC[2][], int pBufferADC[2][])
{
	if (m_pActiveBufferDAC == (((int *) pBufferDAC) + 2*nBufferSize))
		return DSR0 - ((int *) pBufferDAC);
	
	return DSR0 - (((int *) pBufferDAC) + 2*nBufferSize);
}
