/*******************************************************************************
 * C H A M E L E O N   DSP C file                                              *
 *******************************************************************************
 * Copyright (C) 2001-2002 Soundart                                            *
 * www.soundart-hot.com                                                        *
 * support@soundart-hot.com                                                    *
 ******************************************************************************/

#include <stdlib.h>
#include <dsp56303.h>

int m_nHostBufferSize;
int m_nHostBufferMask;

int * m_pHostTxDataBufferEnd;
int * volatile m_pHostTxDataWrite;
int * volatile m_pHostTxDataRead;

int * m_pHostRxDataBufferEnd;
int * volatile m_pHostRxDataWrite;
int * volatile m_pHostRxDataRead;

void ConfigHostInterrupts(int nBufferSize, int *pBufferHostTx, int *pBufferHostRx)
{
	/**********************************************************************
	 * Disable Host Port
	 *********************************************************************/
	HPCR.B.HEN = 0;

	/**********************************************************************
	 * Disable Host Interrupts
	 *********************************************************************/
	HCR.B.HRIE = 0;
	HCR.B.HTIE = 0;
	HCR.B.HCIE = 0;
	 
	/**********************************************************************
	 * Initialize private data
	 *********************************************************************/
	m_pHostRxDataBufferEnd = pBufferHostRx + nBufferSize;
	m_pHostRxDataWrite = pBufferHostRx;
	m_pHostRxDataRead  = pBufferHostRx;
	
	m_pHostTxDataBufferEnd = pBufferHostTx + nBufferSize;
	m_pHostTxDataWrite = pBufferHostTx;
	m_pHostTxDataRead  = pBufferHostTx;

	m_nHostBufferSize = nBufferSize;
	m_nHostBufferMask = nBufferSize - 1;

	/**********************************************************************
	 * Enable Host Interrupts
	 *********************************************************************/
	HCR.B.HRIE = 1;

	/**********************************************************************
	 * Enable Host Port
	 *********************************************************************/
	HPCR.B.HEN = 1;
}

void WriteHostData(int nDataSize, int *pData)
{
	int *pHostTxDataWrite = m_pHostTxDataWrite;
	int i;
	
	for (i=0; i<nDataSize ;i++)
	{
		pHostTxDataWrite++;
		if (pHostTxDataWrite >= m_pHostTxDataBufferEnd)
			pHostTxDataWrite -= m_nHostBufferSize;

		while (pHostTxDataWrite == m_pHostTxDataRead)
			HCR.B.HTIE = 1;

		*m_pHostTxDataWrite++ = *pData++;
		if (m_pHostTxDataWrite >= m_pHostTxDataBufferEnd)
			m_pHostTxDataWrite -= m_nHostBufferSize;
	}

	HCR.B.HTIE = 1;
}

void ReadHostData(int nDataSize, int *pData)
{
	while (nDataSize--)
	{
		*pData++ = *m_pHostRxDataRead++;
		if (m_pHostRxDataRead >= m_pHostRxDataBufferEnd)
			m_pHostRxDataRead -= m_nHostBufferSize;
	}
	HCR.B.HRIE = 1;
}

int *GetHostReceivedDataPointer(void)
{
	return m_pHostRxDataRead;
}

int GetHostReceivedDataCount(void)
{
	int nCount;

	if (m_pHostRxDataWrite == m_pHostRxDataRead)
		return 0;

	nCount = m_pHostRxDataWrite - m_pHostRxDataRead;
	if (nCount < 0)
		nCount += m_nHostBufferSize;

	return nCount;
}

