/* heavily based upon soundart's g563c example */

/*******************************************************************************
 * C H A M E L E O N   ColdFire C file                                         *
 *******************************************************************************
 * Copyright (C) 2001-2003 Soundart                                            *
 * www.soundart-hot.com                                                        *
 * support@soundart-hot.com                                                    *
 ******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <rtems.h>
#include <chameleon.h>
#include <string.h>
//#include <midishare.h>

#include "dsp/dsp_code.h"
#include "ihm.h"


#define VER_MAJ 0
#define VER_MIN 9


/******************************************************************************/
#define WORKSPACE_SIZE	128*1024
rtems_unsigned32 rtems_workspace_size = WORKSPACE_SIZE;
rtems_unsigned32 rtems_workspace_start[WORKSPACE_SIZE];
/******************************************************************************/

rtems_id dsp_mutex,midi_mutex,midi_task;

#define dsp_lock() rtems_semaphore_obtain(dsp_mutex,RTEMS_WAIT,100)
#define dsp_unlock() rtems_semaphore_release(dsp_mutex)

#define midi_lock() rtems_semaphore_obtain(dsp_mutex,RTEMS_WAIT,100)
#define midi_unlock() rtems_semaphore_release(dsp_mutex)


MidiEvPtr CLOCK_EV;

extern int mainpage_init;

void Load_Config();
void _Save_Config(char*,unsigned int);


void dsp_write(rtems_unsigned32 word);

char*param_names[N_PARAMETERS]= {
	"Left In Volume", "Right In Volume",
	"L1 Out Volume", "L2 Out Volume", "L3 Out Volume", "L4 Out Volume",
	"R1 Out Volume", "R2 Out Volume", "R3 Out Volume", "R4 Out Volume",
	"Left Thru Gain", "Right Thru Gain" };


rtems_unsigned32 db2coef[128] = {
        0, 0x91, 0x9E, 0xAC, 0xBC, 0xCD, 0xE0, 0xF4,
        0x10A, 0x122, 0x13D, 0x15A, 0x179, 0x19C, 0x1C1, 0x1EA,
        0x217, 0x247, 0x27C, 0x2B6, 0x2F5, 0x33A, 0x385, 0x3D7,
        0x431, 0x493, 0x4FD, 0x571, 0x5F0, 0x67A, 0x711, 0x7B5,
        0x869, 0x92D, 0xA02, 0xAEB, 0xBEA, 0xCFF, 0xE2E, 0xF78,
        0x10E0, 0x1268, 0x1415, 0x15E9, 0x17E7, 0x1A13, 0x1C72, 0x1F08,
        0x21DA, 0x24EE, 0x284A, 0x2BF4, 0x2FF3, 0x344F, 0x3911, 0x3E41,
        0x43EB, 0x4A17, 0x50D4, 0x582E, 0x6032, 0x68F1, 0x727C, 0x7CE5,
        0x8840, 0x94A4, 0xA228, 0xB0E6, 0xC0FC, 0xD288, 0xE5AD, 0xFA8F,
        0x11158, 0x12A32, 0x1454F, 0x162E4, 0x18329, 0x1A65C, 0x1CCC4, 0x1F6A9,
        0x2245E, 0x2563A, 0x28C9F, 0x2C7F6, 0x308B3, 0x34F52, 0x39C5D, 0x3F069,
        0x44C1A, 0x4B022, 0x51D41, 0x5944D, 0x6162B, 0x6A3D9, 0x73E68, 0x7E706,
        0x89EF8, 0x967A4, 0xA428F, 0xB3161, 0xC35EB, 0xD5225, 0xE8837, 0xFDA7A,
        0x114B80, 0x12DE13, 0x149541, 0x16745D, 0x187F0A, 0x1AB93D, 0x1D2749, 0x1FCDE4,
        0x22B234, 0x25D9D3, 0x294AE1, 0x2D0C0C, 0x31249C, 0x359C83, 0x3A7C71, 0x3FCDD3,
        0x459AFE, 0x4BEF33, 0x52D6BA, 0x5A5EF8, 0x629689, 0x6B8D5F, 0x7554DF, 0x7FFFFF
};


int panel;
int dsp;

int looper_state[2][4];

unsigned int loop_mem_size=0;
unsigned int loop_mem_base=0;

#define DSPSYNC 0
#define DSPIM 1

rtems_unsigned32 dsp_fifo[2][128];
int dsp_fifo_r[2]={0,0}, dsp_fifo_w[2]={0,0};

#define dsp_fifo_can_write(_n) (dsp_fifo_r[_n]!=((dsp_fifo_w[_n]+1)&127))
#define dsp_fifo_can_read(_n) (dsp_fifo_r[_n]!=dsp_fifo_w[_n])

#ifndef DEBUG
#define DEFAULT_PRESET \
	{ { { { 60,61,62,63 }, { 64,65,66,67 }, { 68,69,70,71 }, { 72,73,74,75 } },\
	  { { 76,77,78,79 }, { 80,81,82,83 }, { 84,85,86,87 }, { 88,89,90,91 } } },\
	  { 1,2, 3,4,5,6, 7,8,9,10, 11,12 },\
	  { 127,127, 127,127,127,127, 127,127,127,127, 127,127 },\
	  -1, 24, 0, 0 }
#else
#define DEFAULT_PRESET \
	{ { { { 115,117,119,120 }, { 64,65,66,67 }, { 68,69,70,71 }, { 72,73,74,75 } },\
	  { { 76,77,78,79 }, { 80,81,82,83 }, { 84,85,86,87 }, { 88,89,90,91 } } },\
	  { 1,2, 3,4,5,6, 7,8,9,10, 11,12 },\
	  { 127,127, 127,127,127,127, 127,127,127,127, 127,127 },\
	  -1, 24, 0, 0 }
#endif

#define EMPTY_PRESET \
	{ { { { 0,0,0,0 }, { 0,0,0,0 }, { 0,0,0,0 }, { 0,0,0,0 } },\
	  { { 0,0,0,0 }, { 0,0,0,0 }, { 0,0,0,0 }, { 0,0,0,0 } } },\
	  { 0,0, 0,0,0,0, 0,0,0,0, 0,0 },\
	  { 127,127, 127,127,127,127, 127,127,127,127, 127,127 },\
	  -1, 0, 0, 0 }

Config_t config;
Config_t default_conf = {
//	rtems_unsigned32 magic;
	CONF_MAGIC,
	
//	char p;
	0,

//	char presetnames[N_PRESETS][NAMESIZE+1];
	{ "Default ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ",
	  "Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   ","Empty   " },
	  
//	Preset_t P[N_PRESETS];
	{ DEFAULT_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET,
	  EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET, EMPTY_PRESET }
};









#include <EventToMidiStream.h>
#include <MidiStreamToEvent.h>

static Ev2StreamRec xmt;

LinearizeMthTbl	lTbl;

const char* gNoteNames[12] = { "C", "C#", "D", "Eb", "E", "F", "F#", "G", "G#", "A", "Bb", "B" };


extern float loop_length_modes[N_MODES][4];

int looper_on[2][4];

int volatile clock_is_running=0;
int beat_counter=0,tick_counter=0;
unsigned int total_tick=-1;


volatile rtems_interval ledl=0,ledr=0,ledc=0;
int volatile led_state=0;


#define LED_PULSE_LENGTH 15


extern int p_action,p_lr,p_num,p_param;

short refMidi;

int volatile dsp_usage;
int volatile dsp_usage_debug[16];

int no_message=0;

float Volumes[12]= { -144.f, -144.f, -144.f, -144.f, -144.f, -144.f, -144.f, -144.f, -144.f, -144.f, -144.f, -144.f };


typedef struct {
	unsigned int counter,length,state,sync;
} MIDI_Looper_t;

MIDI_Looper_t midi_looper[2][4]= {{{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,0,0,0}},
				  {{0,0,0,0}, {0,0,0,0}, {0,0,0,0}, {0,0,0,0}}};


extern int mode;


void set_parameter(int n,int value);
void set_memmode(int value);
void set_looper_state(int lr,int looper, int state);


void set_channel() {
	int c=Preset.channel;
	TFilter*f=MidiGetFilter(refMidi);
	if(!f) {
		TRACE("\nERROR !!! Couldn't get MIDI filter !!!\n\n");
		return;
	}
	if(c==-1) {
		f->channel[0]=0xFF;
		f->channel[1]=0xFF;
	} else {
		f->channel[0]=0;
		f->channel[1]=0;
		AcceptBit(f->channel,c);
	}
	MidiSetFilter(refMidi,f);
	TRACE("MIDI Channel set to %i\n",c);
	Save_Config(Preset.channel);
}




void SelectPreset(int n) {
	unsigned int v[2];
	int i;
	if(n<0)
		config.p=0;
	else if(n>127)
		config.p=127;
	else
		config.p=n;
	if(!mode)
		display_message("Select Preset   %3.3i:%8.8s    ",config.p,config.presetnames[(int)config.p]);
	no_message=1;
	TRACE("Select Preset #%i\n",n);
	
	for(i=0;i<N_PARAMETERS;i++) {
		set_parameter(i,(int)Preset.volumes[i]);
	}

	set_memmode(Preset.looplength_mode);
	set_channel();
	Save_Config(config.p);
	no_message=0;
}






//	char Action_Assigns[2][4][4]; /* left/right, 4 loopers, 4 actions (play_once play_loop stop record)
//					assign is 1:7 on/off note# */





static __inline rtems_unsigned32 dsp_fifo_read(int n) {	/* only one reader, trivial implementation */
	rtems_unsigned32 ret=dsp_fifo[n][dsp_fifo_r[n]];
	++dsp_fifo_r[n];
	dsp_fifo_r[n]&=0x7F;
	return ret;
}

/*static __inline*/
void dsp_fifo_write(int n,rtems_unsigned32 word) {	/* lock-free and thread-safe implementation */
	int w,z;
	w=dsp_fifo_w[n];
	z=w;
	++z;
	z&=0x7F;
	dsp_fifo_w[n]=z;
	dsp_fifo[n][w]=word;
}



void _Save_Config(char*ptr,unsigned int size) {
#ifndef DEBUG
	int flash=flash_init();
	unsigned int ofs=ptr-(char*)&config;
	if(ofs>=sizeof(Config_t)||(ofs+size)>sizeof(Config_t)) {
		TRACE("ERROR :: trying to save [%p] out of bounds (size %i)\n",ptr,size);
		return;
	}
//	flash_write_data(flash,0,(char*)&config,sizeof(Config_t));
	flash_write_data(flash,ofs,ptr,size);
	flash_exit(flash);
#endif
}



void Load_Config() {
#ifndef DEBUG
	int flash=flash_init();
	flash_read_data(flash,0,(char*)&config,sizeof(Config_t));
	flash_exit(flash);
	if(config.magic!=CONF_MAGIC) {
		TRACE("Using default configuration\n");
		memmove(&config,(char*)&default_conf,sizeof(Config_t));
		_Save_Config((char*)&config,sizeof(Config_t));
	}
#else
	memmove(&config,(char*)&default_conf,sizeof(Config_t));
#endif
}








char*value2ascii(signed int value,char*buf) {
	float f;
	if(value==127) {
		strcpy(buf,"  0.0 dB ");
	} else if(value) {
		f=20.f*log(fix_to_float(db2coef[value]))/log(10.f);
		sprintf(buf,"%4.1f dB",f);
	} else {
		strcpy(buf," -oo dB ");
	}
//	TRACE("value2ascii(%i) => \"%s\"\n",value,buf);
	return buf;
}







void led_pulse_left() {
//	TRACE("led_pulse_left\n");
	disp_lock();
	rtems_clock_get(RTEMS_CLOCK_GET_TICKS_SINCE_BOOT,(rtems_interval*)&ledl);
	ledl+=LED_PULSE_LENGTH;
	led_state|=PANEL01_LED_SHIFT;
	panel_out_led(panel,led_state);
	disp_unlock();
}

void led_pulse_clock() {
//	TRACE("led_pulse_left\n");
	disp_lock();
	rtems_clock_get(RTEMS_CLOCK_GET_TICKS_SINCE_BOOT,(rtems_interval*)&ledc);
	ledc+=LED_PULSE_LENGTH;
	led_state|=PANEL01_LED_EDIT;
	panel_out_led(panel,led_state);
	disp_unlock();
}

void led_pulse_right() {
//	TRACE("led_pulse_right\n");
	if(!clock_is_running) return;
	disp_lock();
	rtems_clock_get(RTEMS_CLOCK_GET_TICKS_SINCE_BOOT,(rtems_interval*)&ledr);
	ledr+=LED_PULSE_LENGTH;
	led_state|=PANEL01_LED_CTRL1|PANEL01_LED_CTRL2|PANEL01_LED_CTRL3;
	panel_out_led(panel,led_state);
	disp_unlock();
}


void do_leds() {
	rtems_interval date;
	//static int old_state=0;
	disp_lock();
	rtems_clock_get(RTEMS_CLOCK_GET_TICKS_SINCE_BOOT,&date);
	if(ledl&&date>=ledl+LED_PULSE_LENGTH) {
		led_state&=~(PANEL01_LED_SHIFT);
		ledl=0;
	}
	if(ledr&&date>=ledr+LED_PULSE_LENGTH) {
		led_state&=~(PANEL01_LED_CTRL1|PANEL01_LED_CTRL2|PANEL01_LED_CTRL3);
		ledr=0;
	}
	if(ledc&&date>=ledc+LED_PULSE_LENGTH) {
		led_state&=~(PANEL01_LED_EDIT);
		ledc=0;
	}
	//if(led_state!=old_state) {
	panel_out_led(panel,led_state);
	//	old_state=led_state;
	//}
	disp_unlock();
}





void set_looper_state(int lr,int looper, int state) {
	MIDI_Looper_t*m=&(midi_looper[lr][looper]);
	int k;
	/*
		- si on dmarre un enregistrement, counter=tick_counter
		- si on stoppe  //       //      , length=tick_counter-counter;
		- si on dmarre une lecture, counter=0
		- si on stoppe une lecture, rien de spcial
		- si en lecture, si en boucle et counter==length alors counter=0 et envoie commande PLAY|LOOP
	*/
	if(state&STATE_BIT_REC) {		/* RECORD */
//		TRACE("STATE_BIT_REC\n");
		m->counter=tick_counter;
		m->length=0;
		m->sync=clock_is_running;
		TRACE("looper start record T=%i s=%i\n",tick_counter,clock_is_running);
	} else {				/* STOP */
//		TRACE("MIDI: m->newstate&0 (STOP)\n");
		if(m->state&STATE_BIT_REC) {
			TRACE("looper stop record T=%i s=%i\n",tick_counter,m->sync);
			if(tick_counter>m->counter)
				m->length=tick_counter-m->counter;
			else
				m->length=m->counter-tick_counter;
			m->sync=(m->sync&&clock_is_running);
			TRACE("       stop record L=%i s=%i\n",m->length,m->sync);
		}
		m->counter=-1;
	}

	m->state=state;
	k=0xFFFF00|MkAssign(lr,looper,state);
	/*if(!tick_counter) {*/
		/* FIXME : finir d'implmenter le send immdiat des commandes ct CF (+ envoi msg sync) et DSP */
		dsp_write_data(1,&k);
	/*} else {
		dsp_fifo_write(DSPSYNC,k);
	}*/
	
}




void set_parameter(int n,int value) {
	static char dbbuf[10];
	dsp_fifo_write(DSPIM,MkMessage(MSG_BIT_SET,n));
	dsp_fifo_write(DSPIM,value);
	if(!(mode||no_message)) display_message("%-16.16sset to %-11.11s",param_names[n],value2ascii(Preset.volumes[n],dbbuf));
}


void set_memmode(int value) {
	dsp_fifo_write(DSPIM,MkMessage(MSG_BIT_SET|65,value));
}



TFilter filter;



void do_loop_sync() {
	/* FIXME : buggy */
	static unsigned int old_tick_counter=0;
	int k,j;
	int delta=tick_counter-old_tick_counter;
	MIDI_Looper_t*m;

	if(delta) {
		old_tick_counter=tick_counter;
		if(delta>1) TRACE("MIDI: HEY ! clock tick delta = %i !!\n",delta);
//		else TRACE("MIDI:Delta tick #%i\n",tick_counter);
		for(k=0;k<2;k++) for(j=0;j<4;j++) {
			m=&(midi_looper[k][j]);
//			if(!(k+j)) TRACE("m-> s=%i l=%i c=%i s=%i\n",m->sync,m->length,m->counter,m->state);
			if(m->sync&&m->length&&(m->state&STATE_BIT_PLAY)) {
				m->counter+=delta;
				if(m->counter>=m->length) {
					if(m->state&STATE_BIT_LOOP) {
						m->counter%=m->length;
						TRACE("MIDI:retrig looping sample on clock (length=%i error=%i length=%i tick=%i)\n",m->length,m->counter,m->length,tick_counter);
						set_looper_state(k,j,STATE_BIT_PLAY|STATE_BIT_LOOP);
					} else {
						TRACE("MIDI:stop looping sample on clock (length=%i error=%i length=%i tick=%i)\n",m->length,m->counter-m->length,m->length,tick_counter);
						set_looper_state(k,j,0);
					}
				}
			}
		}
	}
}



void send_cmd_sync() {
	int value;
	if(dsp_fifo_can_read(DSPSYNC)&&!(beat_counter&&clock_is_running)) {
		while(dsp_fifo_can_read(DSPSYNC)) {
			value=dsp_fifo_read(DSPSYNC);
			if (!dsp_write_data(dsp,&value,1))
				TRACE("ERROR: can't write to DSP\n");
		}
	}
	while(dsp_fifo_can_read(DSPIM)) {
		value=dsp_fifo_read(DSPIM);
		TRACE("DSPIM 0x%X\n",value);
		if (!dsp_write_data(dsp,&value,1))
			TRACE("ERROR: can't write to DSP\n");
	}
}




static void receive_alarm(short ref) {
	/* si MIDI CLOCK : incr tick_counter; tick_counter==0 ? dsp_send(FFFFFF) : tick_counter==1 ? dsp_send(FFFFFE) : rien */
	midi_loopback();
}






static rtems_task midi_task_routine(rtems_task_argument argument)
{
	//TRACE("R+\n");
	int ass_i,assign,ctrlr,value,i,j,k;
	MidiEvPtr		ev;

	rtems_event_set		pending=-1;
	rtems_status_code	status;
	rtems_id		task_id;
//	float f;
	TRACE("Starting midi_task...\n");
	MidiStreamInit(&xmt, lTbl);
	refMidi = MidiOpen("Midi");
	if(refMidi < 0) {
		TRACE("ERROR: cannot open MidiShare.\n");
		rtems_task_delete(RTEMS_SELF);
	}
	rtems_task_ident(RTEMS_SELF, 0, &task_id);
	CLOCK_EV=MidiNewEv(typeClock);
	MidiSetInfo(refMidi, (void *) task_id);
	MidiSetRcvAlarm(refMidi, receive_alarm);
	AcceptBit(filter.port,0);
	AcceptBit(filter.evType,typeClock);
	AcceptBit(filter.evType,typeProgChange);
	AcceptBit(filter.evType,typeStart);
	AcceptBit(filter.evType,typeStop);
	AcceptBit(filter.evType,typeContinue);
	AcceptBit(filter.evType,typeCtrlChange);
	AcceptBit(filter.evType,typeKeyOn);
	AcceptBit(filter.evType,typeKeyOff);
	filter.channel[0]=0xFF;
	filter.channel[1]=0xFF;
	MidiSetFilter(refMidi,&filter);
	MidiConnect(refMidi, refMidi, TRUE);
	MidiConnect(0, refMidi, TRUE);
	midi_unlock();

	while(TRUE) {
		status = rtems_event_receive(
			EVENT_MIDI|EVENT_NOTE|EVENT_PRG|EVENT_CTRL,
			RTEMS_WAIT | RTEMS_EVENT_ANY,
			RTEMS_NO_TIMEOUT,
			&pending
		);
		if (status != RTEMS_SUCCESSFUL)
			break;

//		TRACE("prout\n");
		while((ev = MidiGetEv(refMidi))) {
			ass_i=0;
			switch(EvType(ev)) {
				case typeKeyOff:
					//TRACE("typeKeyOff,");
					ass_i=128;
				case typeKeyOn:
					led_pulse_left();
					//TRACE("typeKeyOn %li %li\n",MidiGetField(ev,0),MidiGetField(ev,1));
					if(MidiGetField(ev,1)==0)
						ass_i=128;
					//assign=config.NoteEv[ass_i][MidiGetField(ev,0)];
					ass_i+=MidiGetField(ev,0);
					assign=-1;
					for(k=0;k<2;k++) 
					for(j=0;j<4;j++)
					if(looper_on[k][j])
					for(i=0;i<4;i++)
						if(ass_i==Preset.Action_Assigns[k][j][i]) {
							switch(i) {
							case 0: set_looper_state(k,j,STATE_BIT_PLAY); break;
							case 1: set_looper_state(k,j,STATE_BIT_LOOP|STATE_BIT_PLAY); break;
							case 2: set_looper_state(k,j,STATE_BIT_REC); break;
							case 3: set_looper_state(k,j,0);
							};
	//						TRACE("=> envoi d'tat vers le looper %c%i : 0x%X\n",k?'R':'L',j+1,assign);
						}
					//if(assign!=-1) {
					//	dsp_fifo_write(assign);
					//}
					if(midi_learn==1) {
						switch(p_action) {
						case 0: i=STATE_BIT_PLAY; break;
						case 1: i=STATE_BIT_LOOP|STATE_BIT_PLAY; break;
						case 2: i=STATE_BIT_REC; break;
						case 3: i=0;
						};
						*midi_learn_target=(ass_i&128)|MidiGetField(ev,0);
						midi_learn=0;
	//					TRACE("MIDI Note set\n");
						Save_Config(*midi_learn_target);
					}
					break;
				case typeProgChange:
					//TRACE("typeProgChange\n");
					SelectPreset(MidiGetField(ev,0));
					break;
				case typeCtrlChange:
					led_pulse_left();
					//TRACE("typeCtrlChange\n");
					ctrlr=MidiGetField(ev,0);
					ass_i=MidiGetField(ev,1);
					value=db2coef[ass_i];
					if(midi_learn==2) {
						*midi_learn_target=ctrlr;
						midi_learn=0;
	//					TRACE("MIDI Controller set\n");
						Save_Config(*midi_learn_target);
					}
					for(i=0;i<N_PARAMETERS;i++) {
						//TRACE("param %i ctrlr %i :: %i\n",i,config.parameter_controllers[i],ctrlr);
						if(Preset.parameter_controllers[i]==ctrlr) {
							Preset.volumes[i]=ass_i;
							set_parameter(i,value);
	//						dsp_fifo_write(MSG_BIT_SET|(i*256));
	//						dsp_fifo_write(value);
							//TRACE("param #%i change, msg %X %X\n",i,MSG_BIT_SET|(i*256),value);
	/*							if(value) {
								f=20.f*log(fix_to_float(value))/log(10.f);
								display_message("%-16.16sset to %4.1f dB",param_names[i],f);
								mainpage_init=0;
							} else {
								display_message("%-16.16sset to -oo dB",param_names[i]);
								mainpage_init=0;
							}*/
	//							mainpage_init=0;
						}
					}
					break;
				case typeStart:
					TRACE("typeStart");
					total_tick=-1;
				case typeContinue:
					TRACE("typeContinue");
					led_pulse_clock();
					clock_is_running=1;
				case typeClock:
					//TRACE("typeClock\n");
					if(clock_is_running) {
						led_pulse_clock();
						++tick_counter;
						++total_tick;
						if(Preset.beat_counter_max) {
							total_tick%=288;/* 12*24 */
							beat_counter=total_tick;
							beat_counter%=Preset.beat_counter_max;
							if(!beat_counter)
								led_pulse_right();
							//TRACE("BEAT_COUNTER %2.2i MAX %2.2i\n",beat_counter,Preset.beat_counter_max);
						}
					}
					break;
				case typeStop:
					TRACE("typeStop\n");
					led_pulse_clock();
					clock_is_running=0;
					break;
			};
			MidiFreeEv(ev);
			//display_1_event(ev);
		}
		do_loop_sync();
		send_cmd_sync();
		//TRACE("R-\n");
	}

	MidiConnect(0, refMidi, FALSE);

	MidiClose(refMidi);

	/*do rtems_task_wake_after(31337); while(1);*/

	TRACE("Ending midi_task...\n");
	rtems_task_delete(RTEMS_SELF);
}



rtems_boolean create_midi_task(void)
{
	rtems_status_code	status;

	status=rtems_semaphore_create(rtems_build_name('M','I','D','M'),1,
					0,0,&midi_mutex);

	midi_lock();

	MidiStreamInitMthTbl(lTbl);

	status = rtems_task_create(
		rtems_build_name('M', 'I', 'D', 'I'),
		3,
		RTEMS_MINIMUM_STACK_SIZE,
		RTEMS_DEFAULT_MODES,
		RTEMS_DEFAULT_ATTRIBUTES,
		&midi_task
	);
	if (status != RTEMS_SUCCESSFUL)
	{
		TRACE("ERROR: can't create midi_task.\n");
		return FALSE;
	}

	status = rtems_task_start(midi_task, midi_task_routine, 0);
	if (status != RTEMS_SUCCESSFUL)
	{
		TRACE("ERROR: can't start midi_task.\n");
		return FALSE;
	}

	return TRUE;
}




/******************************************************************************/

static rtems_task dsp_read_task(rtems_task_argument argument)
{
	int msg,num,i,l,r,flag=0;
	int dsp = (int) argument;	
	rtems_signed32 data;

	i=rtems_semaphore_create(rtems_build_name('D','S','P','M'),1,
				0,0,&dsp_mutex);

	TRACE("Starting dsp_read_task...\n");

	dsp_lock();

	while (TRUE)
	{
		if (!dsp_read_data(dsp, &data, 1))
			continue;

		if(!(data||flag)) {	/* dsp init reply */
			dsp_unlock();
			flag=1;
		}
		if ((data & 0xFF0000) == 0xFF0000) {
			char ch = (char) (data & 0x7F);
			
			if (ch == 0x0A)
				TRACE("\n");
			else if (ch >= 0x20)
				TRACE("%c", (char) ch);
		} else if ((data & 0xFF0000) == 0xFE0000) {
			r=(data&0xFF00)>>8;
			//TRACE("DSP usage: %d (%d%%)\n", data & 0xFF,(((data&0xFF)+r)*100)/(2*r));
			/*
			dsp_usage=(((data&0xFF)+r)*100)/(2*r);/*/
			dsp_usage=data&0xFF;//*/
		} else if ((data & 0xFF0000) == 0xF00000) {
			dsp_usage_debug[(data>>8)&0xFF]=data&0xFF;
		} else {
			if(data&MSG_BIT_SET) {
				msg=Msg(data);
				if(msg&64) {
					TRACE("reu 0x%X !!\n",data);
					num=msg&0x3F;
					switch(num) {
					case 0:	/* 0x??40 : get loopmem size */
						while(!dsp_read_data(dsp,&loop_mem_size,1));
						while(!dsp_read_data(dsp,&loop_mem_base,1));
						TRACE("oh, so dsp's loop memory size is %i words long\n",loop_mem_size);
						/*display_message("[%6X] %7i                ",loop_mem_size,loop_mem_base);*/
						break;
					case 1: /* 0x??41 : set looplengths mode */
						Preset.looplength_mode=Param(data);
						l=ModeL(Preset.looplength_mode);
						r=ModeR(Preset.looplength_mode);
						for(i=0;i<4;i++) {
							looper_on[0][i]=(loop_length_modes[l][i]>0.f);
							looper_on[1][i]=(loop_length_modes[r][i]>0.f);
						}
						TRACE("looplen mode %i:%i\n",l,r);
						TRACE("looper_on[0] %i %i %i %i\n",looper_on[0][0],looper_on[0][1],looper_on[0][2],looper_on[0][3]);
						TRACE("looper_on[1] %i %i %i %i\n",looper_on[1][0],looper_on[1][1],looper_on[1][2],looper_on[1][3]);
						Save_Config(Preset.looplength_mode);
						break;
					case 2: /* 0x??42 : set internal freq : 1=48000 0=32000 */
						TRACE("reu 0x%X !!\n",data);
						break;
					};
				} else {
					//while(!dsp_read_data(dsp,&data,1));
					num=Msg(data)&0xF;
					i=Param(data);
					if(i&0x8000) i=0;
					else i<<=8;
					if(i==0x7FFF00) {
						Volumes[num]=0.f;
					} else if(i) {
						Volumes[num]=20.f*log(fix_to_float((Param(data)<<8)))/log(10);
					} else {
						Volumes[num]=-144.f;
					}
					/*TRACE("Reu volume #%i (%i) => %f\n",num,Param(data),Volumes[num]);*/
				}
			} else {
				TRACE("Received %X\n",data);
				looper_state[data&MSG_BIT_RIGHT?1:0][data&3]=i=data&MSG_MASK_STATE;
				if((!i)&&midi_looper[data&MSG_BIT_RIGHT?1:0][data&3].state)
					set_looper_state(data&MSG_BIT_RIGHT?1:0,data&3,0);
				TRACE("Now looper_state[%i][%i]=%X\n",data&MSG_BIT_RIGHT?1:0,data&3,looper_state[data&MSG_BIT_RIGHT?1:0][data&3]);
			}
		}
	}

	TRACE("Ending dsp_read_task...\n");
	rtems_task_delete(RTEMS_SELF);
}




rtems_boolean create_dsp_read_task(int dsp)
{
	rtems_id		task_id;
	rtems_status_code	status;

	status = rtems_task_create(
		rtems_build_name('D', 'S', 'P', 'R'),
		10,
		RTEMS_MINIMUM_STACK_SIZE,
		RTEMS_DEFAULT_MODES,
		RTEMS_DEFAULT_ATTRIBUTES,
		&task_id
	);

	if (status != RTEMS_SUCCESSFUL)
	{
		TRACE("ERROR: cannot create dsp_task.\n");
		return FALSE;
	}

	status = rtems_task_start(task_id, dsp_read_task, (rtems_task_argument) dsp);
	if (status != RTEMS_SUCCESSFUL)
	{
		TRACE("ERROR: cannot start dsp_read_task.\n");
		return FALSE;
	}

	return TRUE;
}







extern rtems_interval TMR_ticks_per_s;

/*
void scroll_msg(char*msg,rtems_interval duration) {
	char buf[64];
	int l=strlen(msg);
	int i;
	rtems_interval interval=duration/(16+l);
	sprintf(buf,"                %s                ",msg);
	clear();
	redraw();
	l+=16;
	for(i=0;i<l;i++) {
		panel_out_lcd_print(panel,0,0,buf+i);
		rtems_task_wake_after(interval);
	}
		
}
*/

rtems_task rtems_main(rtems_task_argument ignored)
{
	int i;
	float dB;
	unsigned int v[4];

	TRACE("Magic = 0x%X\n",CONF_MAGIC);

	Load_Config();
	
	dsp = dsp_init(1, dspCode);
	if (!dsp)
	{
		TRACE("ERROR: cannot access the DSP.\n");
		rtems_task_delete(RTEMS_SELF);
	}

	create_dsp_read_task(dsp);
	//create_dsp_write_task(dsp);
	
	/* Fill volume_table with a variable gain from -60 dB to 0 dB */
	for (i=0; i<128 ;i++)
	{
		if (i < 27)
			dB = -90.0f + (float) 40.0f*i/27.0f;
		else
			dB = -50.0f + (float) 50.0f*(i-27)/100.0f;

		db2coef[i] = float_to_fix_round(powf(10.0f, dB/20.0f));
	}

	db2coef[0] = 0;

	create_ihm_task();

	raw_set_chr(1,BOLD_BITMAP('J'));
	raw_set_chr(2,BOLD_BITMAP('L'));
	display_message("     \1erash           \2abs      ");
	
	rtems_task_wake_after(TMR_ticks_per_s);

	raw_set_chr(1,BOLD_BITMAP('L'));
	raw_set_chr(2,BOLD_BITMAP('2'));
	raw_set_chr(4,BOLD_BITMAP('8'));
//	scroll_msg("\1ooper \2\4",TMR_ticks_per_s*4);
	display_message("  \1ooper\2\4\2      version %i.%i      ",VER_MAJ,VER_MIN);
	rtems_task_wake_after(TMR_ticks_per_s);
	v[0]=0;
	/* wait for dsp to init */
	while(dsp_lock()!=RTEMS_SUCCESSFUL) dsp_write_data(dsp,v,1);
	rtems_task_wake_after(10);
	v[0]=MkMessage(MSG_BIT_SET|64,0);
	dsp_write_data(dsp,v,1);

	rtems_task_wake_after(10);
//	create_led_task();

	create_midi_task();

	while(midi_lock()!=RTEMS_SUCCESSFUL) rtems_task_wake_after(10);
	
	SelectPreset(config.p);

	start_ihm_task();

	rtems_task_delete(RTEMS_SELF);
}
